/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#ifndef CAOSDB_UTILS_H
#define CAOSDB_UTILS_H
#include <cassert>
#include <iostream>
#include <string_view>
#include <fstream>
#include <string>
#include <cstdlib>
#include <boost/json.hpp>
#include <boost/beast/core/detail/base64.hpp>
#include <boost/filesystem.hpp>
#include <boost/filesystem/fstream.hpp>

namespace caosdb::utility {
using boost::filesystem::exists;
using boost::filesystem::ifstream;
using boost::filesystem::path;
using boost::json::stream_parser;
using boost::json::value;

/**
 * @brief Read a text file into a string and return the file's content.
 *
 * TODO use boost-filesystem's "load_string_file"!
 */
inline auto load_string_file(const std::string &path) -> std::string {
  const auto path_view = std::string_view{path};
  constexpr auto size = std::size_t{4096};
  auto stream = std::ifstream{path_view.data()};
  stream.exceptions(std::ios_base::badbit);

  auto result = std::string();
  auto buffer = std::string(size, '\0');
  while (stream.read(&buffer[0], size)) {
    result.append(buffer, 0, stream.gcount());
  }
  result.append(buffer, 0, stream.gcount());
  return result;
}

inline auto get_env_var(const char *key, const char *fall_back) -> const
  char * {
  const char *val = getenv(key);
  if (val == nullptr) {
    return fall_back;
  } else {
    return val;
  }
}

/**
 * @brief Return the value of an environment variable or - if undefined - the
 * fall_back value.
 */
inline auto get_env_var(const std::string &key, const std::string &fall_back)
  -> const std::string {
  const char *val = get_env_var(key.c_str(), fall_back.c_str());

  auto const result = std::string(val);
  return result;
}

/**
 * @brief Encode string as base64
 */
inline auto base64_encode(const std::string &plain) -> std::string {
  auto size_plain = plain.size();
  auto size_encoded = boost::beast::detail::base64::encoded_size(size_plain);

  char encoded[size_encoded];
  boost::beast::detail::base64::encode(&encoded, plain.c_str(), size_plain);

  // the encoded char[] is not null terminated, so explicitely set the length
  return std::string(encoded, encoded + size_encoded);
}

inline auto load_json_file(const path &json_file) -> value {
  assert(exists(json_file));

  constexpr auto buffer_size = std::size_t(4096);
  auto stream = ifstream(json_file);
  stream.exceptions(std::ios_base::badbit);

  stream_parser parser;
  auto result = std::string();
  auto buffer = std::string(buffer_size, '\0');
  while (stream.read(&buffer[0], buffer_size)) {
    parser.write(buffer.c_str(), stream.gcount());
  }
  parser.write(buffer.c_str(), stream.gcount());

  assert(parser.done());
  return parser.release();
}

} // namespace caosdb::utility
#endif
