/*
 *
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/connection.h"
#include <grpcpp/create_channel.h>              // for CreateChannel
#include <grpcpp/impl/codegen/client_context.h> // for ClientContext
#include <grpcpp/impl/codegen/status.h>         // for Status
#include <grpcpp/security/credentials.h>        // for SslCredentials
#include <iostream>                             // for operator<<, basic_...
#include <stdexcept>                            // for runtime_error
#include <string>                               // for operator+, char_tr...
#include <memory>
#include "caosdb/authentication.h"                // for Authenticator
#include "caosdb/exceptions.h"                    // for AuthenticationError
#include "caosdb/info/v1alpha1/main.grpc.pb.h"    // for GeneralInfoService
#include "caosdb/info/v1alpha1/main.pb.h"         // for GetVersionInfoResp...
#include "caosdb/transaction.h"                   // for Transaction
#include "caosdb/utility.h"                       // for load_string_file
#include "caosdb/info.h"                          // for VersionInfo
#include "grpcpp/impl/codegen/status_code_enum.h" // for StatusCode, UNAUTH...

namespace caosdb::connection {
using caosdb::authentication::Authenticator;
using caosdb::entity::v1alpha1::EntityTransactionService;
using caosdb::exceptions::AuthenticationError;
using caosdb::exceptions::ConnectionError;
using caosdb::info::VersionInfo;
using caosdb::info::v1alpha1::GeneralInfoService;
using caosdb::info::v1alpha1::GetVersionInfoRequest;
using caosdb::info::v1alpha1::GetVersionInfoResponse;
using caosdb::transaction::Transaction;
using caosdb::utility::load_string_file;
using grpc::InsecureChannelCredentials;
using grpc::SslCredentials;
using grpc::SslCredentialsOptions;

PemFileCACertProvider::PemFileCACertProvider(const std::string &path) {
  this->cacert = load_string_file(path);
}

auto PemFileCACertProvider::GetCACertPem() const -> std::string {
  return this->cacert;
}

PemCACertProvider::PemCACertProvider(const std::string &cacert) {
  this->cacert = cacert;
}

auto PemCACertProvider::GetCACertPem() const -> std::string {
  return this->cacert;
}

CaosDBConnectionConfig::CaosDBConnectionConfig(const std::string &host,
                                               int port) {
  this->host = host;
  this->port = port;
}

auto CaosDBConnectionConfig::GetHost() const -> std::string {
  return this->host;
}

auto CaosDBConnectionConfig::GetPort() const -> int { return this->port; }

auto operator<<(std::ostream &out, const CaosDBConnectionConfig &config)
  -> std::ostream & {
  out << config.ToString();
  return out;
}

InsecureCaosDBConnectionConfig::InsecureCaosDBConnectionConfig(
  const std::string &host, int port)
  : CaosDBConnectionConfig(host, port) {
  this->credentials = InsecureChannelCredentials();
}

auto InsecureCaosDBConnectionConfig::GetChannelCredentials() const
  -> std::shared_ptr<ChannelCredentials> {
  return this->credentials;
}

auto InsecureCaosDBConnectionConfig::ToString() const -> std::string {
  return "InsecureCaosDBConnectionConfig(" + this->GetHost() + "," +
         std::to_string(this->GetPort()) + ")";
}

SslCaosDBConnectionConfig::SslCaosDBConnectionConfig(const std::string &host,
                                                     int port)
  : CaosDBConnectionConfig(host, port) {
  SslCredentialsOptions options;
  this->credentials = SslCredentials(options);
}

SslCaosDBConnectionConfig::SslCaosDBConnectionConfig(
  const std::string &host, int port, const CACertificateProvider &cacert)
  : CaosDBConnectionConfig(host, port) {
  SslCredentialsOptions options;
  options.pem_root_certs = cacert.GetCACertPem();
  this->credentials = SslCredentials(options);
}

SslCaosDBConnectionConfig::SslCaosDBConnectionConfig(
  const std::string &host, int port, const Authenticator &authenticator)
  : CaosDBConnectionConfig(host, port) {

  SslCredentialsOptions options;
  this->credentials = grpc::CompositeChannelCredentials(
    SslCredentials(options), authenticator.GetCallCredentials());
}

SslCaosDBConnectionConfig::SslCaosDBConnectionConfig(
  const std::string &host, int port, const CACertificateProvider &cacert,
  const Authenticator &authenticator)
  : CaosDBConnectionConfig(host, port) {

  SslCredentialsOptions options;
  options.pem_root_certs = cacert.GetCACertPem();
  this->credentials = grpc::CompositeChannelCredentials(
    SslCredentials(options), authenticator.GetCallCredentials());
}

auto SslCaosDBConnectionConfig::GetChannelCredentials() const
  -> std::shared_ptr<ChannelCredentials> {
  return this->credentials;
}

auto SslCaosDBConnectionConfig::ToString() const -> std::string {
  return "SslCaosDBConnectionConfig(" + this->GetHost() + "," +
         std::to_string(this->GetPort()) + "," + this->cacert + ")";
}

CaosDBConnection::CaosDBConnection(const CaosDBConnectionConfig &config) {
  const std::string target =
    config.GetHost() + ":" + std::to_string(config.GetPort());
  this->channel = grpc::CreateChannel(target, config.GetChannelCredentials());
  this->general_info_service = GeneralInfoService::NewStub(this->channel);
  this->entity_transaction_service =
    std::make_shared<EntityTransactionService::Stub>(this->channel);
}

auto operator<<(std::ostream &out, const CaosDBConnection & /*connection*/)
  -> std::ostream & {
  out << "CaosDBConnection()";
  return out;
}

[[nodiscard]] auto CaosDBConnection::GetVersionInfo() const
  -> std::unique_ptr<VersionInfo> {
  const GetVersionInfoRequest request;
  GetVersionInfoResponse response;
  grpc::ClientContext context;
  const grpc::Status status =
    this->general_info_service->GetVersionInfo(&context, request, &response);

  if (!status.ok()) {
    switch (status.error_code()) {
    case grpc::StatusCode::UNAUTHENTICATED:
      throw AuthenticationError(status.error_message());
    case grpc::StatusCode::UNAVAILABLE:
      throw ConnectionError(status.error_message());
    default:
      throw std::runtime_error("Status Code " +
                               std::to_string(status.error_code()) + " - " +
                               status.error_message());
    }
  }
  return std::make_unique<VersionInfo>(response.release_version_info());
}

[[nodiscard]] auto CaosDBConnection::CreateTransaction() const
  -> std::unique_ptr<Transaction> {
  auto service_stub = this->entity_transaction_service;
  return std::make_unique<Transaction>(service_stub);
};

} // namespace caosdb::connection
